"""
OUTFILE FORMAT STRINGS:
  %%              Literal percent sign
  %f              Path to input video file, but without its extension
  %d              Path to directory containing the input video file
  %[W]n           Index of the extracted run, starting from 0
  %[W]N           Index of the extracted run, starting from 1
  %t    / %T      Timestamp of first/last frame, in timedelta format (H:MM:SS)
  %[W]s / %[W]S   Timestamp of first/last frame, in seconds since start
  %[W]m / %[W]M   Timestamp of first/last frame, in ms since start
  %[W]i / %[W]I   Frame number of first/last frame, with optional pad width

[W] above indicates an optional numeric width to zero-pad the attribute to.
"""

from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict

from .core import Slide


@dataclass
class Fmter:
    attr: Callable[[Slide], Any]
    widthable: bool = False


TAGS: Dict[str, Fmter] = dict(
    # can't use % as kwarg, so put it separately
    [("%", Fmter(lambda _: "%"))],
    # use kwargs for the rest so Python can catch duplicates
    # (https://stackoverflow.com/a/22242922)
    f=Fmter(lambda s: s.source),
    d=Fmter(lambda s: s.source.parent),
    n=Fmter(lambda s: s.index, True),
    N=Fmter(lambda s: s.index0, True),
    t=Fmter(lambda s: s.start.timestamp),
    s=Fmter(lambda s: s.start.timestamp.total_seconds(), True),
    m=Fmter(lambda s: s.start.timestamp.total_seconds() * 1000, True),
    i=Fmter(lambda s: s.start.fnum, True),
    T=Fmter(lambda s: s.end.timestamp),
    S=Fmter(lambda s: s.end.timestamp.total_seconds(), True),
    M=Fmter(lambda s: s.end.timestamp.total_seconds() * 1000, True),
    I=Fmter(lambda s: s.end.fnum, True),
)


@dataclass
class OutputPathFormatter:
    slide: Slide
    outfmt: str

    _OUTFMT_RE: ClassVar[re.Pattern] = re.compile(
        r"""
            (?<!%)          # can't be preceded by a % sign
            %               # literal % to start the format tag
            (?P<width>\d*)  # optional width qualifier (ignored if not applicable)
            (?P<char>.)     # match any character so we can complain about unescaped
                            # percents or bad tags
        """,
        re.VERBOSE,
    )

    def _outfmt_repl(self, match: re.Match) -> Any:
        width, char = match.groups()
        tag = TAGS[char]
        out = str(tag.attr(self.slide))
        if tag.widthable and width:
            out = out.zfill(int(width))
        return out

    def output_path(self) -> Path:
        return self._OUTFMT_RE.sub(self._outfmt_repl, self.outfmt)



"""
# video-slide-extractor

Finds runs of consecutive similar frames from a video file, and exports first and/or
last frames from each run as images. Useful for e.g. extracting slides from a recorded
presentation.

Lists exported image file paths on stdout.


Usage:
  {program} <video_file>
            [-b <beg_outfmt>] [-e <end_outfmt>]
            [-s <sample_rate>] [-c <corr>]
            [-d | --debug]

  {program} (-h | --help)
  {program} (-v | --version)


OPTIONS:

  -b, --begfmt <string>
        OUTFILE FORMAT STRING to determine filename for first frame in each run
        [default: none]

  -e, --endfmt <string>
        OUTFILE FORMAT STRING to determine filename for last frame in each run
        [default: %f__%t.png]

  -s, --sample-rate ( <float|int>fps | <float|int>% )
        approximate frequency with which to sample video frames to check for changes.
        must be suffixed with "fps" or "%". an fps value is divided into the source
        video's native framerate in FPS to determine # frames to skip each iteration. a
        percentage value is interpreted as a percentage of the source video's native
        framerate and handled similarly. [default: 1.0fps]

  -c, --correlation-threshold <float>
        correlation threshold above which two frames are considered the same
        [default: 0.999]

  -d, --debug
        drop into an ipdb debugger for every run found; requires [debug] extra

  -h, --help
        print help

  -v, --version
        print version


{outfile_format_docs}


Inspired by https://github.com/szanni/slideextract.
"""

from functools import partial
from pathlib import Path
from sys import argv
from typing import Any

import cv2 as cv
import numpy as np
from docopt import docopt
from single_source import get_version
from tqdm import tqdm

from .core import VideoSlideExtractor, frames_match
from .output_path_format import OutputPathFormatter
from .output_path_format import __doc__ as outfile_format_docs


def main():
    version: str = str(get_version(__name__, Path(__file__).parent))
    program: str = Path(argv[0]).name
    doc: str = str(__doc__).format(
        program=program, outfile_format_docs=str(outfile_format_docs).strip()
    )
    args: dict[str, Any] = docopt(doc=doc, version=version)

    matcher = partial(frames_match, float(args["--correlation-threshold"]))
    extractor = VideoSlideExtractor(
        Path(args["<video_file>"]), args["--sample-rate"], matcher
    )

    progress = tqdm(total=int(extractor.total_frames()), unit=" frames")
    extractor._on_update = progress.update

    for slide in extractor.slides():
        if args["--debug"]:
            np.set_printoptions(edgeitems=0)
            import ipdb  # fmt: skip
            ipdb.set_trace()

        images = {"--begfmt": slide.start.image, "--endfmt": slide.end.image}
        for argname, image in images.items():
            if (outfmt := args[argname]) and outfmt != "none":
                fmter = OutputPathFormatter(slide, outfmt)
                path = fmter.output_path()
                progress.clear()
                print(path)
                progress.refresh()
                cv.imwrite(str(path), image)

    progress.close()


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import timedelta
from functools import cached_property
from pathlib import Path
from typing import Callable, Iterator, TypeAlias

import cv2 as cv
from more_itertools import first, last, split_when
from numpy.typing import NDArray

Image: TypeAlias = NDArray


FrameMatcher = Callable[["Frame", "Frame"], bool]


def frames_match(correlation_threshold: float, frame1: Frame, frame2: Frame) -> bool:
    match = cv.matchTemplate(frame1.image, frame2.image, method=cv.TM_CCORR_NORMED)
    return (match > correlation_threshold).all()  # type: ignore


@dataclass
class VideoSlideExtractor:
    path: Path
    sample_rate: str
    frame_matcher: FrameMatcher
    _capture: cv.VideoCapture = field(init=False)
    _on_update: Callable = lambda: ...

    def __post_init__(self):
        self._capture = cv.VideoCapture(str(self.path))

    def frame_rate(self) -> int:
        return self._capture.get(cv.CAP_PROP_FPS)

    @cached_property
    def frame_sample_period(self) -> int:
        if self.sample_rate and self.sample_rate.endswith("fps"):
            return int(self.frame_rate() / float(self.sample_rate[:-3]))
        elif self.sample_rate and self.sample_rate.endswith("%"):
            return int(self.frame_rate() * float(self.sample_rate[:-1]))
        else:
            return int(self.total_frames() / (int(self.sample_rate) - 1))

    def current_frame(self) -> Frame:
        success, image = self._capture.retrieve()
        assert success
        return Frame(
            image=image,
            fnum=self.current_frame_num(),
            timestamp=self.current_timestamp(),
            source=self.path,
        )

    def current_frame_num(self) -> int:
        return self._capture.get(cv.CAP_PROP_POS_FRAMES)

    def total_frames(self) -> int:
        return self._capture.get(cv.CAP_PROP_FRAME_COUNT)

    def current_timestamp(self) -> timedelta:
        return timedelta(milliseconds=self._capture.get(cv.CAP_PROP_POS_MSEC))

    def frames(self) -> Iterator[Frame]:
        fnum = 0
        while self._capture.grab():
            if fnum % self.frame_sample_period == 0:
                yield self.current_frame()
            fnum += 1
            self._on_update()

    def frame_runs(self) -> Iterator[list[Frame]]:
        return split_when(self.frames(), lambda *a: not self.frame_matcher(*a))

    def slides(self) -> Iterator[Slide]:
        return (
            Slide(i, first(frame_run), last(frame_run))
            for i, frame_run in enumerate(self.frame_runs())
        )


@dataclass
class Frame:
    image: Image
    fnum: int
    timestamp: timedelta
    source: Path


@dataclass
class Slide:
    index: int
    start: Frame
    end: Frame

    @property
    def index0(self) -> int:
        return self.index - 1

    @cached_property
    def source(self) -> Path:
        assert self.start.source == self.end.source
        return self.start.source



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['video_slide_extractor']

package_data = \
{'': ['*']}

install_requires = \
['docopt>=0.6.2,<0.7.0',
 'more-itertools>=8.12.0,<9.0.0',
 'opencv-python>=4.5.5,<5.0.0',
 'single-source>=0.3.0,<0.4.0',
 'tqdm>=4.63.0,<5.0.0']

extras_require = \
{'debug': ['ipdb>=0.13.9,<0.14.0']}

entry_points = \
{'console_scripts': ['slide-extractor = video_slide_extractor:cli.main']}

setup_kwargs = {
    'name': 'video-slide-extractor',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Tim Vergenz',
    'author_email': 'vergenzt@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
